'$DYNAMIC
'Huffman Compressor
'compresses a screen 13(VGA) picture
'Rich Geldreich
'December 24, 1991
'You may use or modify this program in any way you wish.

DEFINT A-Z
DECLARE SUB SendByte (B)
DECLARE SUB SendBit (Bit)
DECLARE SUB Deletenode (A)
DECLARE FUNCTION EmtyNode ()

'Make the tree as easy as store as possible
'Typ- the type of node
'NextLeft- the left child
'NextRight- the right child
'Contents- whatever it holds- it could be a number or letter(byte)
'root- TRUE if at the top
'Divs-
'RealIndex- only used when cleaning up the tree for transmission
TYPE Nodetype
    Typ AS INTEGER
    NextLeft AS INTEGER
    NextRight AS INTEGER
    Contents AS LONG
    Root AS INTEGER
    Divs AS INTEGER
    RealIndex AS INTEGER
END TYPE

'declare some usefull constants
CONST True = -1, False = NOT True
CONST NotUsed = -1, Byte = 1, Number = 2
CONST Alone = 1, Mult = 2

DIM Count(255) AS LONG
DIM Direction(50) AS INTEGER
DIM Flowest AS LONG, Nlowest AS LONG
DIM Bytes(256), Legal(600)

DIM SHARED Bitsin AS INTEGER, TempChar  AS INTEGER
DIM SHARED Powers(8) AS INTEGER, Exps(9) AS INTEGER
DIM SHARED Unuse, Roots
DIM SHARED Position AS INTEGER

REDIM SHARED Unused(600) AS INTEGER
REDIM SHARED Node(600) AS Nodetype
REDIM Codes(256, 64) AS INTEGER, Numcodes(256) AS INTEGER
REDIM SHARED ByteBuffer(32700) AS STRING * 2

'set up the bit twiddling numbers
FOR A = 1 TO 8: Powers(A) = 2 ^ (8 - A): NEXT
FOR A = 9 TO 0 STEP -1: Exps(A) = 2 ^ A: NEXT

'clear the tree
FOR A = 0 TO 600
    Node(A).Typ = NotUsed
    Node(A).NextLeft = NotUsed
    Node(A).NextRight = NotUsed
    Node(A).Contents = 0
    Node(A).Root = False
    Node(A).Divs = Alone
NEXT

'print my stupid title
PRINT "An Example of Huffman Compression"
PRINT "By Rich Geldreich 1991"
PRINT "Press any key to proceed."
A$ = INPUT$(1)

'go to VGA
RANDOMIZE TIMER '<--like it's going to make a difference in my cheap example!

SCREEN 13, , 0, 0
DEF SEG = &HA000

'make some graphics
FOR A = 1 TO 400
    IF RND > .05 THEN
        LINE -(RND * 320, RND * 200), RND * 255
    ELSE
        LINE -(RND * 320, RND * 200), RND * 255, BF
    END IF
NEXT

'find out how much each color is used and store so it can
'be put into a tree
FOR X = 0 TO 319
    FOR Y = 0 TO 199
        A = POINT(X, Y): Count(A) = Count(A) + 1
    NEXT
NEXT



'Unuse and Unused() make up a system to find emty nodes fast and easy
'instead of scanning the Node() array; when a node is deleted it gets
'put on the top of Unused()- much like a stack of some sorts
Unuse = 600
FOR A = 600 TO 0 STEP -1: Unused(600 - A) = A: NEXT

'Roots tells how many nodes are at the top
Roots = 0
FOR A = 0 TO 255
    'lets make a node if a char is used
    IF Count(A) <> 0 THEN
       
        'find an emty node
        B = EmtyNode
        'this one holds the number
        Node(B).Typ = Number
        Node(B).Contents = Count(A)
        Node(B).Root = True: Roots = Roots + 1
        Node(B).Divs = Alone
       
        'find another emty node
        B1 = EmtyNode
        'this one holds the letter: the first node created above
        '"points" to this one through NextLeft
        Node(B1).Typ = Byte
        Node(B1).Contents = A
        Node(B1).Root = False
        Node(B1).NextLeft = NotUsed
        Node(B1).NextRight = NotUsed
        Node(B1).Divs = Alone

        'let the number point to the letter
        Node(B).NextLeft = B1
        Node(B).NextRight = NotUsed

    END IF
NEXT


'when Roots=1, then the tree is finished because there is
'only one node at the top
DO UNTIL Roots = 1
    'scan for the two lowest nodes
    'this could of been done faster- but it works
    
    Flowest = 999999: Fpointer = -1

    FOR A = 0 TO 600
        IF Node(A).Root AND Node(A).Contents <= Flowest THEN
            Npointer = Fpointer: Flowest = Node(A).Contents: Fpointer = A
        END IF
    NEXT
    IF Npointer = -1 THEN
        Nlowest = 999999
        FOR A = 0 TO 600
            IF Node(A).Root AND A <> Fpointer AND Node(A).Contents < Nlowest THEN
                Npointer = A: Nlowest = Node(A).Contents
            END IF
        NEXT
    END IF

    'lets combine the two
    A = EmtyNode
    Node(A).Typ = Number
    'add the counts
    Node(A).Contents = Node(Fpointer).Contents + Node(Npointer).Contents
    Node(A).Divs = Mult
    'put it on the top
    Node(A).Root = True: Roots = Roots + 1

    'take care of putting these nodes together; a number doesn't point
    'to another number unless the other number points to more than one
    'letter or number!
    IF Node(Fpointer).Divs = Alone AND Node(Npointer).Divs = Mult THEN
        Node(A).NextLeft = Node(Fpointer).NextLeft
        Node(A).NextRight = Npointer
        Node(Npointer).Root = False: Roots = Roots - 1
        Deletenode Fpointer
   
    ELSEIF Node(Fpointer).Divs = Mult AND Node(Npointer).Divs = Alone THEN
        Node(A).NextLeft = Node(Npointer).NextLeft
        Node(A).NextRight = Fpointer
        Node(Fpointer).Root = False: Roots = Roots - 1
        Deletenode Npointer
   
    ELSEIF Node(Fpointer).Divs = Mult AND Node(Npointer).Divs = Mult THEN
        Node(A).NextLeft = Npointer
        Node(A).NextRight = Fpointer
        Node(Npointer).Root = False
        Node(Fpointer).Root = False: Roots = Roots - 2
   
    ELSEIF Node(Fpointer).Divs = Alone AND Node(Npointer).Divs = Alone THEN
        Node(A).NextLeft = Node(Fpointer).NextLeft
        Node(A).NextRight = Node(Npointer).NextLeft
        Deletenode Fpointer
        Deletenode Npointer
   
    END IF
LOOP
'the tree is ready


'Find all the characters in the tree for quick access
FOR A = 0 TO 600
    IF Node(A).Typ = Byte THEN
        Bytes(Node(A).Contents) = A
    END IF
NEXT

'Find all the numbers in the tree
Uses = 0
FOR A = 0 TO 600
    IF Node(A).Typ = Number THEN
        Legal(Uses) = A: Uses = Uses + 1
    END IF
NEXT: Uses = Uses - 1

'start scanning the tree and build up the codes
'note that we can't go down the tree- we got to go up it
'so out destination is the very top
'as we go up it we record whether we take the left or right path
'
FOR A = 0 TO 255
    'find the codes if the character is used
    IF Count(A) <> 0 THEN
        'find out character in the tree and start looking there
        B = Bytes(A)
        'Num records how many codes for this char
        Num = 0
        'start going up
        DO
            'look for this nodes parent
            Left = False: Right = False
            FOR C1 = 0 TO Uses
                C = Legal(C1)
                IF Node(C).NextLeft = B THEN
                    Direct = 0: EXIT FOR
                ELSEIF Node(C).NextRight = B THEN
                    Direct = 1: EXIT FOR
                END IF
            NEXT
            'we found the parent and made part of a code
            Direction(Num) = Direct
            Num = Num + 1
            'reposition ourselves with this new found node and
            'continue until we reach the top
            B = C
        LOOP UNTIL Node(B).Root = True
    ELSE
        Num = 0
    END IF
    Numcodes(A) = Num
    C = 0
    'switch everything around because we put all the
    'codes in backwards
    FOR B = Num - 1 TO 0 STEP -1
        Codes(A, C) = Direction(B)
        C = C + 1
    NEXT
NEXT


'now we must send the tree to the output file
'the tree must be cleaned up so it can be sent as
'small as possible
'the tree is reindexed(much like GW-BASIC's RENUM command)
'so only the nodes that are part of the tree are included
'otherwise we would have to send all 600 nodes- that's too much!
Position = 0
Bitsin = 0
TempChar = 0
Index = 0
'find all the nodes in the tree and tag them with their new numbers
FOR A = 0 TO 600
    IF Node(A).Typ <> NotUsed THEN
        Node(A).RealIndex = Index
        Index = Index + 1
    END IF
NEXT: Index = Index - 1
're-adjust the NextLeft and NextRight pointers to point
'twards the new tags
FOR A = 0 TO 600
    IF Node(A).Typ <> NotUsed THEN
        IF Node(A).NextLeft <> NotUsed THEN
            Node(A).NextLeft = Node(Node(A).NextLeft).RealIndex
        ELSE
            Node(A).NextLeft = 1023
        END IF
        IF Node(A).NextRight <> NotUsed THEN
            Node(A).NextRight = Node(Node(A).NextRight).RealIndex
        ELSE
            Node(A).NextRight = 1023
        END IF
    END IF
NEXT
'find the poor guy at the very top...
FOR A = 0 TO 600
    IF Node(A).Root THEN
        EXIT FOR
    END IF
NEXT
RootIndex = Node(A).RealIndex

'start sending (note that the codes are just put in the ByteBuffer()
'array and the array is stored to disk later)

'send the number of nodes and the node at the top
SendByte Index
SendByte RootIndex

'send the tree; only send what is needed
FOR A = 0 TO 600
    IF Node(A).Typ <> NotUsed THEN
        SendByte Node(A).NextLeft
        SendByte Node(A).NextRight
       
        'send a character; but don't send a number it isn't needed
        IF Node(A).Typ = Byte THEN
            SendBit 1
            FOR B = 1 TO 8
                IF (Node(A).Contents AND Powers(B)) > 0 THEN
                    SendBit 1
                ELSE
                    SendBit 0
                END IF
            NEXT
        ELSE
            SendBit 0
        END IF
    END IF
NEXT


'Wow! Were almost done! Now we start compression...
'Scan the screen and send the codes that stand for each pixel's color
FOR X = 0 TO 319
    FOR Y = 0 TO 199
        'get color
        Bte = POINT(X, Y)
        'send the codes
        FOR B = 0 TO Numcodes(Bte) - 1
            SendBit Codes(Bte, B)
        NEXT
    NEXT
    'clear a line to report progress
    LINE (X, 0)-(X, 199), 0
NEXT

'this is a dirty way of making sure we have stored all the
'bits in the buffer
FOR A = 1 TO 16: SendBit 0: NEXT

SCREEN 0: WIDTH 80
'report compression and store file to disk
PRINT "Compression:"; INT(((64000 - CLNG(Position) * 2) / 65536) * 100 + .5); "%"
DEF SEG = VARSEG(ByteBuffer(0))
BSAVE "screen.huf", VARPTR(ByteBuffer(0)), CLNG(Position) * 2
'all done
BEEP
END

REM $STATIC
'this sub deletes a node from the tree and puts it back
'in free space
SUB Deletenode (A)
    Node(A).Typ = NotUsed
    Node(A).NextLeft = NotUsed
    Node(A).NextRight = NotUsed
    Node(A).Contents = 0
    IF Node(A).Root = True THEN Roots = Roots - 1: Node(A).Root = False
    Node(A).Divs = Alone
    Unuse = Unuse + 1
    Unused(Unuse) = A
END SUB

'finds an emty node and takes it away from free space
FUNCTION EmtyNode
    EmtyNode = Unused(Unuse): Unuse = Unuse - 1
END FUNCTION

'stores a single bit in buffer
SUB SendBit (Bit) STATIC
    Bitsin = Bitsin + 1
    IF Bitsin = 9 THEN
        A$ = A$ + CHR$(TempChar)
        BytesIn = BytesIn + 1
        IF BytesIn = 2 THEN
            ByteBuffer(Position) = A$
            A$ = ""
            BytesIn = 0
            Position = Position + 1
        END IF
        Bitsin = 1: TempChar = 0
    END IF
    TempChar = (TempChar * 2) + Bit
END SUB

'sends a 10 bits in buffer
SUB SendByte (B)
    FOR A = 9 TO 0 STEP -1
        IF (B AND Exps(A)) > 0 THEN
            SendBit 1
        ELSE
            SendBit 0
        END IF
    NEXT
END SUB

